You are viewing a plain text version of this content. The canonical link for it is here.
Posted to commits@jackrabbit.apache.org by st...@apache.org on 2005/08/19 18:42:09 UTC

svn commit: r233511 - in /incubator/jackrabbit/trunk/core/src: java/org/apache/jackrabbit/core/ java/org/apache/jackrabbit/core/state/ test/org/apache/jackrabbit/core/

Author: stefan
Date: Fri Aug 19 09:41:57 2005
New Revision: 233511

URL: http://svn.apache.org/viewcvs?rev=233511&view=rev
Log:
fixed concurrency issue reported by walter raboch:
http://www.mail-archive.com/jackrabbit-dev@incubator.apache.org/msg02127.html

Added:
    incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/StaleItemStateException.java   (with props)
    incubator/jackrabbit/trunk/core/src/test/org/apache/jackrabbit/core/ConcurrencyTest.java   (with props)
Modified:
    incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/ItemImpl.java
    incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/SessionImpl.java
    incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/ChangeLog.java
    incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/ItemState.java
    incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/LocalItemStateManager.java
    incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/SharedItemStateManager.java
    incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/TransactionalItemStateManager.java
    incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/UpdatableItemStateManager.java
    incubator/jackrabbit/trunk/core/src/test/org/apache/jackrabbit/core/TestAll.java

Modified: incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/ItemImpl.java
URL: http://svn.apache.org/viewcvs/incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/ItemImpl.java?rev=233511&r1=233510&r2=233511&view=diff
==============================================================================
--- incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/ItemImpl.java (original)
+++ incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/ItemImpl.java Fri Aug 19 09:41:57 2005
@@ -33,6 +33,7 @@
 import org.apache.jackrabbit.core.state.NodeState;
 import org.apache.jackrabbit.core.state.PropertyState;
 import org.apache.jackrabbit.core.state.SessionItemStateManager;
+import org.apache.jackrabbit.core.state.StaleItemStateException;
 import org.apache.jackrabbit.core.value.InternalValue;
 import org.apache.jackrabbit.core.version.VersionManager;
 import org.apache.jackrabbit.name.MalformedPathException;
@@ -1306,7 +1307,6 @@
             Collection dirtyRefs =
                     checkReferences(dirty.iterator(), removed.iterator());
 
-
             // start the update operation
             try {
                 stateMgr.edit();
@@ -1358,6 +1358,8 @@
                 stateMgr.update();
                 // update operation succeeded
                 succeeded = true;
+            } catch (StaleItemStateException e) {
+                throw new InvalidItemStateException(e.getMessage());
             } catch (ItemStateException e) {
                 String msg = safeGetJCRPath() + ": unable to update item.";
                 log.debug(msg);

Modified: incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/SessionImpl.java
URL: http://svn.apache.org/viewcvs/incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/SessionImpl.java?rev=233511&r1=233510&r2=233511&view=diff
==============================================================================
--- incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/SessionImpl.java (original)
+++ incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/SessionImpl.java Fri Aug 19 09:41:57 2005
@@ -1099,14 +1099,12 @@
         // notify listeners that session is about to be closed
         notifyLoggingOut();
 
-        // discard all transient changes
-        itemStateMgr.disposeAllTransientItemStates();
+        // dispose session item state manager
+        itemStateMgr.dispose();
         // dispose item manager
         itemMgr.dispose();
         // dispose workspace
         wsp.dispose();
-
-        // @todo release session-scoped locks, free resources, prepare to get gc'ed etc.
 
         // invalidate session
         alive = false;

Modified: incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/ChangeLog.java
URL: http://svn.apache.org/viewcvs/incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/ChangeLog.java?rev=233511&r1=233510&r2=233511&view=diff
==============================================================================
--- incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/ChangeLog.java (original)
+++ incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/ChangeLog.java Fri Aug 19 09:41:57 2005
@@ -232,6 +232,10 @@
         while (iter.hasNext()) {
             ItemState state = (ItemState) iter.next();
             state.setStatus(ItemState.STATUS_EXISTING);
+            if (!state.hasOverlayedState()) {
+                // update modification count
+                state.touch();
+            }
             state.notifyStateUpdated();
         }
         iter = deletedStates();

Modified: incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/ItemState.java
URL: http://svn.apache.org/viewcvs/incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/ItemState.java?rev=233511&r1=233510&r2=233511&view=diff
==============================================================================
--- incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/LocalItemStateManager.java (original)
+++ incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/LocalItemStateManager.java Fri Aug 19 09:41:57 2005
@@ -306,7 +306,9 @@
     /**
      * {@inheritDoc}
      */
-    public void update() throws ItemStateException, IllegalStateException {
+    public void update()
+            throws StaleItemStateException, ItemStateException,
+            IllegalStateException {
         if (!editMode) {
             throw new IllegalStateException("Not in edit mode");
         }
@@ -323,9 +325,12 @@
      * items with our copies.
      *
      * @param changeLog change log containing local states and references
+     * @throws StaleItemStateException if at least one of the affected item
+     *                                 states has become stale in the meantime
+     * @throws ItemStateException if an error occurs
      */
     protected void update(ChangeLog changeLog)
-            throws ItemStateException {
+            throws StaleItemStateException, ItemStateException {
 
         ObservationManagerImpl obsMgr = null;
 

Modified: incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/SharedItemStateManager.java
URL: http://svn.apache.org/viewcvs/incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/SharedItemStateManager.java?rev=233511&r1=233510&r2=233511&view=diff
==============================================================================
--- incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/SharedItemStateManager.java (original)
+++ incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/SharedItemStateManager.java Fri Aug 19 09:41:57 2005
@@ -74,8 +74,8 @@
     /**
      * Virtual item state providers
      */
-    private VirtualItemStateProvider[] virtualProviders = new
-            VirtualItemStateProvider[0];
+    private VirtualItemStateProvider[] virtualProviders =
+            new VirtualItemStateProvider[0];
 
     /**
      * Read-/Write-Lock to synchronize access on this item state manager.
@@ -320,10 +320,12 @@
      * @param local  change log containing local items
      * @param obsMgr the observation manager to inform, or <code>null</code> if
      *               no observation manager should be informed.
-     * @throws ItemStateException if an error occurs
+     * @throws StaleItemStateException if at least one of the affected item
+     *                                 states has become stale
+     * @throws ItemStateException if another error occurs
      */
     public void store(ChangeLog local, ObservationManagerImpl obsMgr)
-            throws ItemStateException {
+            throws StaleItemStateException, ItemStateException {
 
         ChangeLog shared = new ChangeLog();
 
@@ -376,12 +378,22 @@
                 while (iter.hasNext()) {
                     ItemState state = (ItemState) iter.next();
                     state.connect(getItemState(state.getId()));
+                    if (state.isStale()) {
+                        String msg = state.getId() + " has been modified externally";
+                        log.debug(msg);
+                        throw new StaleItemStateException(msg);
+                    }
                     shared.modified(state.getOverlayedState());
                 }
                 iter = local.deletedStates();
                 while (iter.hasNext()) {
                     ItemState state = (ItemState) iter.next();
                     state.connect(getItemState(state.getId()));
+                    if (state.isStale()) {
+                        String msg = state.getId() + " has been modified externally";
+                        log.debug(msg);
+                        throw new StaleItemStateException(msg);
+                    }
                     shared.deleted(state.getOverlayedState());
                 }
                 iter = local.addedStates();
@@ -628,11 +640,15 @@
     private ItemState loadItemState(ItemId id)
             throws NoSuchItemStateException, ItemStateException {
 
+        ItemState state;
         if (id.denotesNode()) {
-            return persistMgr.load((NodeId) id);
+            state = persistMgr.load((NodeId) id);
         } else {
-            return persistMgr.load((PropertyId) id);
+            state = persistMgr.load((PropertyId) id);
         }
+        // init modification counter
+        state.touch();
+        return state;
     }
 
     /**

Added: incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/StaleItemStateException.java
URL: http://svn.apache.org/viewcvs/incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/StaleItemStateException.java?rev=233511&view=auto
==============================================================================
--- incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/StaleItemStateException.java (added)
+++ incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/StaleItemStateException.java Fri Aug 19 09:41:57 2005
@@ -0,0 +1,66 @@
+/*
+ * Copyright 2004-2005 The Apache Software Foundation or its licensors,
+ *                     as applicable.
+ *
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
+ *
+ *      http://www.apache.org/licenses/LICENSE-2.0
+ *
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
+ */
+package org.apache.jackrabbit.core.state;
+
+import org.apache.jackrabbit.BaseException;
+
+/**
+ * Signals that an item has been modified externally and that the item state
+ * representing it has thus become stale.
+ */
+public class StaleItemStateException extends ItemStateException {
+
+    /**
+     * Constructs a new instance of this class with <code>null</code> as its
+     * detail message.
+     */
+    public StaleItemStateException() {
+        super();
+    }
+
+    /**
+     * Constructs a new instance of this class with the specified detail
+     * message.
+     *
+     * @param message the detail message. The detail message is saved for
+     *                later retrieval by the {@link #getMessage()} method.
+     */
+    public StaleItemStateException(String message) {
+        super(message);
+    }
+
+    /**
+     * Constructs a new instance of this class with the specified detail
+     * message and root cause.
+     *
+     * @param message   the detail message. The detail message is saved for
+     *                  later retrieval by the {@link #getMessage()} method.
+     * @param rootCause root failure cause
+     */
+    public StaleItemStateException(String message, Throwable rootCause) {
+        super(message, rootCause);
+    }
+
+    /**
+     * Constructs a new instance of this class with the specified root cause.
+     *
+     * @param rootCause root failure cause
+     */
+    public StaleItemStateException(Throwable rootCause) {
+        super(rootCause);
+    }
+}

Propchange: incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/StaleItemStateException.java
------------------------------------------------------------------------------
    svn:eol-style = native

Modified: incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/TransactionalItemStateManager.java
URL: http://svn.apache.org/viewcvs/incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/TransactionalItemStateManager.java?rev=233511&r1=233510&r2=233511&view=diff
==============================================================================
--- incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/TransactionalItemStateManager.java (original)
+++ incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/TransactionalItemStateManager.java Fri Aug 19 09:41:57 2005
@@ -257,7 +257,8 @@
      * the ones already known (removing items that were first added and
      * then again deleted).
      */
-    protected void update(ChangeLog changeLog) throws ItemStateException {
+    protected void update(ChangeLog changeLog)
+            throws StaleItemStateException, ItemStateException {
         if (txLog != null) {
             txLog.merge(changeLog);
         } else {

Modified: incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/UpdatableItemStateManager.java
URL: http://svn.apache.org/viewcvs/incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/UpdatableItemStateManager.java?rev=233511&r1=233510&r2=233511&view=diff
==============================================================================
--- incubator/jackrabbit/trunk/core/src/test/org/apache/jackrabbit/core/ConcurrencyTest.java (added)
+++ incubator/jackrabbit/trunk/core/src/test/org/apache/jackrabbit/core/ConcurrencyTest.java Fri Aug 19 09:41:57 2005
@@ -0,0 +1,156 @@
+/*
+ * Copyright 2004-2005 The Apache Software Foundation or its licensors,
+ *                     as applicable.
+ *
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
+ *
+ *      http://www.apache.org/licenses/LICENSE-2.0
+ *
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
+ */
+package org.apache.jackrabbit.core;
+
+import org.apache.jackrabbit.test.AbstractJCRTest;
+
+import javax.jcr.InvalidItemStateException;
+import javax.jcr.Node;
+import javax.jcr.Session;
+import java.util.Random;
+import java.util.ArrayList;
+import java.util.Iterator;
+
+public class ConcurrencyTest extends AbstractJCRTest {
+
+    private static final int NUM_ITERATIONS = 2;
+    private static final int NUM_SESSIONS = 5;
+    private static final int NUM_NODES = 100;
+
+    final ArrayList exceptions = new ArrayList();
+
+    protected void setUp() throws Exception {
+        super.setUp();
+        // @todo setup test environment
+    }
+
+    protected void tearDown() throws Exception {
+        try {
+            // @todo cleanup test environment
+        } finally {
+            super.tearDown();
+        }
+    }
+
+    /**
+     * Runs the test.
+     */
+    public void testConcurrentWritingSessions() throws Exception {
+        int n = NUM_ITERATIONS;
+        while (n-- > 0) {
+            Thread[] threads = new Thread[NUM_SESSIONS];
+            for (int i = 0; i < threads.length; i++) {
+                // create new session
+                Session session = helper.getSuperuserSession();
+                TestSession ts = new TestSession("s" + i, session);
+                Thread t = new Thread(ts);
+                t.setName("s" + i);
+                t.start();
+                threads[i] = t;
+                Thread.sleep(100);
+            }
+            for (int i = 0; i < threads.length; i++) {
+                threads[i].join();
+            }
+        }
+
+        if (!exceptions.isEmpty()) {
+            Exception e = null;
+            for (Iterator it = exceptions.iterator(); it.hasNext();) {
+                e = (Exception) it.next();
+                e.printStackTrace(log);
+            }
+            throw e;
+            //fail();
+        }
+    }
+
+    //--------------------------------------------------------< inner classes >
+    class TestSession implements Runnable {
+
+        Session session;
+        String identity;
+        Random r;
+
+        TestSession(String identity, Session s) {
+            session = s;
+            this.identity = identity;
+            r = new Random();
+        }
+
+        private void randomSleep() {
+            long l = r.nextInt(900) + 200;
+            try {
+                Thread.sleep(l);
+            } catch (InterruptedException ie) {
+            }
+        }
+
+        public void run() {
+
+            log.println("started.");
+            String state = "";
+            try {
+                Node rn = session.getRootNode().getNode(testPath);
+
+                state = "searching testnode";
+                Node n;
+                try {
+                    if (rn.hasNode("testnode-" + identity)) {
+                        state = "removing testnode";
+                        rn.getNode("testnode-" + identity).remove();
+                        session.save();
+                        randomSleep();
+                    }
+                    state = "adding testnode";
+                    n = rn.addNode("testnode-" + identity, "nt:unstructured");
+                    session.save();
+                } catch (InvalidItemStateException e) {
+                    // expected
+                    log.println("encountered InvalidItemStateException while " + state + ", quitting...");
+                    //e.printStackTrace(log);
+                    return;
+                }
+
+                state = "setting property";
+                n.setProperty("testprop", "Hello World!");
+                session.save();
+                randomSleep();
+
+                for (int i = 0; i < NUM_NODES; i++) {
+                    state = "adding subnode " + i;
+                    n.addNode("x" + i, "nt:unstructured");
+                    state = "adding property to subnode " + i;
+                    n.setProperty("testprop", "xxx");
+                    if (i % 10 == 0) {
+                        state = "saving pending subnodes";
+                        session.save();
+                    }
+                    randomSleep();
+                }
+                session.save();
+            } catch (Exception e) {
+                log.println("Exception while " + state + ": " + e.getMessage());
+                exceptions.add(e);
+            } finally {
+                session.logout();
+            }
+
+            log.println("ended.");
+        }
+    }
+}
\ No newline at end of file

Propchange: incubator/jackrabbit/trunk/core/src/test/org/apache/jackrabbit/core/ConcurrencyTest.java
------------------------------------------------------------------------------
    svn:eol-style = native

Modified: incubator/jackrabbit/trunk/core/src/test/org/apache/jackrabbit/core/TestAll.java
URL: http://svn.apache.org/viewcvs/incubator/jackrabbit/trunk/core/src/test/org/apache/jackrabbit/core/TestAll.java?rev=233511&r1=233510&r2=233511&view=diff
==============================================================================
--- incubator/jackrabbit/trunk/core/src/test/org/apache/jackrabbit/core/TestAll.java (original)
+++ incubator/jackrabbit/trunk/core/src/test/org/apache/jackrabbit/core/TestAll.java Fri Aug 19 09:41:57 2005
@@ -35,6 +35,8 @@
     public static Test suite() {
         TestSuite suite = new TestSuite("Core tests");
 
+        //suite.addTestSuite(ConcurrencyTest.class);
+        //suite.addTestSuite(ConcurrentSaveTest.class);
         suite.addTestSuite(XATest.class);
 
         return suite;
--- incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/ItemState.java (original)
+++ incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/ItemState.java Fri Aug 19 09:41:57 2005
@@ -38,6 +38,8 @@
 
     private static Logger log = Logger.getLogger(ItemState.class);
 
+    private static long LAST_MOD_COUNT = 1;
+
     /**
      * flags defining the current status of this <code>ItemState</code> instance
      */
@@ -78,8 +80,7 @@
      */
     protected String parentUUID;
 
-    protected long lastModified;
-
+    private transient long modCount;
 
     protected ItemId id;
 
@@ -92,7 +93,8 @@
      * Listeners (weak references)
      */
     private final transient Map listeners =
-            Collections.synchronizedMap(new ReferenceMap(ReferenceMap.WEAK, ReferenceMap.WEAK));
+            Collections.synchronizedMap(
+                    new ReferenceMap(ReferenceMap.WEAK, ReferenceMap.WEAK));
 
     // the backing persistent item state (may be null)
     protected transient ItemState overlayedState;
@@ -119,8 +121,7 @@
         }
         this.id = id;
         this.parentUUID = parentUUID;
-        // @todo use modification count instead of ms (not precise enough)
-        lastModified = System.currentTimeMillis();
+        modCount = 0;
         overlayedState = null;
         this.isTransient = isTransient;
     }
@@ -128,7 +129,7 @@
     /**
      * Protected constructor
      *
-     * @param initialStatus  the initial status of the new <code>ItemState</code> instance
+     * @param initialStatus the initial status of the new <code>ItemState</code> instance
      * @param isTransient   flag indicating whether this state is transient or not
      */
     protected ItemState(int initialStatus, boolean isTransient) {
@@ -152,16 +153,32 @@
      */
     protected void copy(ItemState state) {
         parentUUID = state.getParentUUID();
-        lastModified = state.getLastModified();
         id = state.getId();
     }
 
     /**
+     * Returns the modification count.
+     * @return the modification count.
+     */
+    long getModCount() {
+        return modCount;
+    }
+
+    /**
+     * Updates the modification count.
+     */
+    synchronized void touch() {
+        modCount = LAST_MOD_COUNT++;
+    }
+
+    /**
      * Pull state information from overlayed state.
      */
     void pull() {
         if (overlayedState != null) {
             copy(overlayedState);
+            // sync modification count
+            modCount = overlayedState.getModCount();
         }
     }
 
@@ -175,8 +192,16 @@
     }
 
     /**
-     * Called by <code>TransientItemStateManager</code> when this item state
-     * is disposed.
+     * Determines whether this item state has become stale.
+     * @return true if this item state has become stale, false otherwise.
+     */
+    boolean isStale() {
+        return overlayedState != null && modCount != overlayedState.getModCount();
+    }
+
+    /**
+     * Called by <code>TransientItemStateManager</code> and
+     * <code>LocalItemStateManager</code> when this item state has been disposed.
      */
     void onDisposed() {
         // prepare this instance so it can be gc'ed
@@ -401,15 +426,6 @@
     }
 
     /**
-     * Returns the timestamp when this item state was last modified.
-     *
-     * @return the timestamp when this item state was last modified.
-     */
-    public long getLastModified() {
-        return lastModified;
-    }
-
-    /**
      * Add an <code>ItemStateListener</code>
      *
      * @param listener the new listener to be informed on modifications
@@ -436,6 +452,7 @@
     public void stateCreated(ItemState created) {
         // underlying state has been permanently created
         status = STATUS_EXISTING;
+        pull();
     }
 
     /**
@@ -443,7 +460,7 @@
      */
     public void stateDestroyed(ItemState destroyed) {
         // underlying state has been permanently destroyed
-        if (isTransient || status != STATUS_EXISTING) {
+        if (isTransient) {
             status = STATUS_STALE_DESTROYED;
         } else {
             status = STATUS_EXISTING_REMOVED;
@@ -456,7 +473,7 @@
      */
     public void stateModified(ItemState modified) {
         // underlying state has been modified
-        if (isTransient || status != STATUS_EXISTING) {
+        if (isTransient) {
             status = STATUS_STALE_MODIFIED;
         } else {
             // this instance represents existing state, update it

Modified: incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/LocalItemStateManager.java
URL: http://svn.apache.org/viewcvs/incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/LocalItemStateManager.java?rev=233511&r1=233510&r2=233511&view=diff
==============================================================================
--- incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/UpdatableItemStateManager.java (original)
+++ incubator/jackrabbit/trunk/core/src/java/org/apache/jackrabbit/core/state/UpdatableItemStateManager.java Fri Aug 19 09:41:57 2005
@@ -109,10 +109,13 @@
      * added to this update operation in a single step.
      * If this operation fails, no item will have been saved.
      *
      * @throws ItemStateException    if the operation failed
      * @throws IllegalStateException if the manager is not in edit mode.
      */
-    void update() throws ItemStateException, IllegalStateException;
+    void update() throws StaleItemStateException, ItemStateException,
+            IllegalStateException;
 
     /**
      * Disposes this <code>UpdatableItemStateManager</code> and frees resources.
@@ -109,10 +109,13 @@
      * If this operation fails, no item will have been saved.
      *
-     * @throws ItemStateException    if the operation failed
-     * @throws IllegalStateException if the manager is not in edit mode.
+     * @throws StaleItemStateException if at least one of the affected items
+     *                                 has become stale in the meantime 
+     * @throws ItemStateException      if the operation failed for another reason
+     * @throws IllegalStateException   if the manager is not in edit mode.
      */

Added: incubator/jackrabbit/trunk/core/src/test/org/apache/jackrabbit/core/ConcurrencyTest.java
URL: http://svn.apache.org/viewcvs/incubator/jackrabbit/trunk/core/src/test/org/apache/jackrabbit/core/ConcurrencyTest.java?rev=233511&view=auto
==============================================================================